@@ -355,7 +355,7 @@ unsigned char *ath_init_driver(unsigned char *param, unsigned char *display)
 	*/
 	ath_win_y_pos = (ath_y_size - ath_y_panel - ath_win_y_size) / 2 + ath_y_panel;
 	ath_win_x_pos = ath_x_size - ath_win_x_size - ath_win_y_pos;
-	if (/*ath_cs_desktop == CS_RGB32 ||*/ ath_cs_desktop == CS_RGB24 || ath_cs_desktop == CS_RGB16 || ath_cs_desktop == CS_RGB15) 
+	if (/*ath_cs_desktop == CS_RGB32 ||*/ ath_cs_desktop == CS_RGB24 || ath_cs_desktop == CS_RGB16 || ath_cs_desktop == CS_RGB15)
 		ath_cs_bmp = ath_cs_desktop;
 	else if (ath_cs_desktop == CS_RGB32 || ath_cs_desktop == CS_RGBA32) ath_cs_bmp = CS_RGB24;
 	else ath_cs_bmp = CS_RGB15;
@@ -520,22 +520,6 @@ void ath_draw_bitmap(struct graphics_device *dev, struct bitmap *bmp, int x, int
 	unlock_dev(dev);
 }
 
-#if 0
-void ath_draw_bitmaps(struct graphics_device *dev, struct bitmap **bmp, int n, int x, int y)
-{
-	LinksView *lvv = lv(dev);
-	lock_dev(dev);
-	while (n--) {
-		Bitmap *b = (Bitmap *)(*bmp)->flags;
-		lvv->DrawBitmap(b, b->GetBounds(), Rect(x, y, x + (*bmp)->x, y + (*bmp)->y));
-		x += (*bmp)->x;
-		bmp++;
-	}
-	lv(dev)->d_flush();
-	unlock_dev(dev);
-}
-#endif
-
 long ath_get_color(int rgb)
 {
 	if (small_color) {
@@ -622,6 +606,12 @@ void ath_set_clip_area(struct graphics_device *dev, struct rect *r)
 	unlock_dev(dev);
 }
 
+void ath_flush(struct graphics_device *dev)
+{
+	unregister_bottom_half(do_flush, dev);
+	do_flush(dev);
+}
+
 struct graphics_driver atheos_driver = {
 	(unsigned char *)"atheos",
 	ath_init_driver,
@@ -635,13 +625,11 @@ struct graphics_driver atheos_driver = {
 	NULL,
 	NULL,
 	ath_get_empty_bitmap,
-	/*ath_get_filled_bitmap,*/
 	ath_register_bitmap,
 	ath_prepare_strip,
 	ath_commit_strip,
 	ath_unregister_bitmap,
 	ath_draw_bitmap,
-	/*ath_draw_bitmaps,*/
 	ath_get_color,
 	ath_fill_area,
 	ath_draw_hline,
@@ -649,6 +637,7 @@ struct graphics_driver atheos_driver = {
 	ath_hscroll,
 	ath_vscroll,
 	ath_set_clip_area,
+	ath_flush,
 	dummy_block,
 	dummy_unblock,
 	ath_set_title,
