@@ -72,29 +72,31 @@ static void grx_set_clip(void)
 		);
 }
 
-static void grx_key_in(void *p, unsigned char *ev_, int size)
+static void grx_key_in(struct itrm *p, unsigned char *ev_, int size)
 {
 	int vd;
 	struct links_event *ev = (struct links_event *)(void *)ev_;
 	if (size != sizeof(struct links_event)) return;
 	if (ev->ev == EV_ABORT) terminate_loop = 1;
 	if (ev->ev != EV_KBD) return;
-	if ((ev->y & KBD_ALT) && ev->x >= '0' && ev->x <= '9') {
+	if (ev->y & KBD_PASTE) goto skip;
+	if (((ev->y & (KBD_CTRL | KBD_ALT)) == KBD_ALT) && ev->x >= '0' && ev->x <= '9') {
 		vd = (ev->x - '1' + 10) % 10;
 		goto switch_vd;
 	}
-	if ((ev->y & KBD_ALT) && ev->x == '-') {
+	if (((ev->y & (KBD_CTRL | KBD_ALT)) == KBD_ALT) && ev->x == '-') {
 		vd = 10;
 		goto switch_vd;
 	}
-	if ((ev->y & KBD_ALT) && ev->x == '=') {
+	if (((ev->y & (KBD_CTRL | KBD_ALT)) == KBD_ALT) && ev->x == '=') {
 		vd = 11;
 		goto switch_vd;
 	}
-	if ((ev->y & KBD_ALT) && ev->x <= KBD_F1 && ev->x >= KBD_F12) {
+	if (((ev->y & (KBD_CTRL | KBD_ALT)) == KBD_ALT) && ev->x <= KBD_F1 && ev->x >= KBD_F12) {
 		vd = KBD_F1 - ev->x;
 		goto switch_vd;
 	}
+skip:
 	if (g_kbd_codepage(&grx_driver) != utf8_table && ev->x >= 128 && ev->x <= 255)
 		if ((ev->x = cp2u(ev->x, g_kbd_codepage(&grx_driver))) == -1) return;
 	if (current_virtual_device && current_virtual_device->keyboard_handler) current_virtual_device->keyboard_handler(current_virtual_device, ev->x, ev->y);
@@ -541,13 +543,11 @@ struct graphics_driver grx_driver = {
 	NULL,
 	NULL,
 	grx_get_empty_bitmap,
-	/*grx_get_filled_bitmap,*/
 	grx_register_bitmap,
 	grx_prepare_strip,
 	grx_commit_strip,
 	grx_unregister_bitmap,
 	grx_draw_bitmap,
-	/*grx_draw_bitmaps,*/
 	NULL,
 	grx_fill_area,
 	grx_draw_hline,
@@ -555,6 +555,7 @@ struct graphics_driver grx_driver = {
 	grx_hscroll,
 	grx_vscroll,
 	grx_set_clip_area,
+	NULL,
 	grx_block,
 	grx_unblock,
 	NULL, /* set_title */
