@@ -32,16 +32,6 @@
 #define NO_IE
 #endif
 
-#define LINKS_MANUAL_URL		(!strcmp(cast_const_char language_name(get_current_language()),"Czech") || !strcmp(cast_const_char language_name(get_current_language()),"Slovak")?\
-					"http://links.twibright.com/user.html":\
-					"http://links.twibright.com/user_en.html")
-#define LINKS_HOMEPAGE_URL		(!strcmp(cast_const_char language_name(get_current_language()),"Czech") || !strcmp(cast_const_char language_name(get_current_language()),"Slovak")?\
-					"http://links.twibright.com/index_cz.php":\
-					"http://links.twibright.com/")
-#define LINKS_CALIBRATION_URL		(!strcmp(cast_const_char language_name(get_current_language()),"Czech") || !strcmp(cast_const_char language_name(get_current_language()),"Slovak")?\
-					"http://links.twibright.com/kalibrace.html":\
-					"http://links.twibright.com/calibration.html")
-
 #define LINKS_SOCK_NAME			"socket"
 #define LINKS_ANONYMOUS_SOCK_SUFFIX	"-a"
 #define LINKS_PORT			23755
@@ -52,11 +42,14 @@
 
 #define FG_POLL_TIME			500
 
+#define OPENMP_MAX_THREADS		8
+
 #define HEAPMIN_FACTOR			100
 #define HEAPMIN_INTERVAL		10000
 #define MALLOC_TRIM_INTERVAL		300000
 
 #define DNS_TIMEOUT			3600000UL
+#define SESSION_TIMEOUT			14400000UL
 
 #define HTTP_KEEPALIVE_TIMEOUT		300000
 #define FTP_KEEPALIVE_TIMEOUT		600000
@@ -174,6 +167,7 @@
 #define DISPLAY_TIME_MAX_FIRST		1000
 #define DISPLAY_TIME			15
 #define IMG_DISPLAY_TIME		7
+#define DISPLAY_FORMATTING_STATUS	500
 
 #define STAT_UPDATE_MIN			100
 #define STAT_UPDATE_MAX			1000
